/**
    Copyright (C) 2002  Michael Ahlberg, Måns Rullgård

    Permission is hereby granted, free of charge, to any person
    obtaining a copy of this software and associated documentation
    files (the "Software"), to deal in the Software without
    restriction, including without limitation the rights to use, copy,
    modify, merge, publish, distribute, sublicense, and/or sell copies
    of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be
    included in all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
    EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
    MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
    NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
    HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
    WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
    DEALINGS IN THE SOFTWARE.
**/

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <semaphore.h>
#include <tcstring.h>
#include <tctypes.h>
#include <pthread.h>
#include <tclist.h>
#include <tchash.h>
#include <sys/time.h>
#include <stddef.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/socket.h>
#include <netdb.h>
#include <openssl/sha.h>
#include <netq_tc2.h>

typedef struct {
    eventq_t q;
    int socket;
    int *scntp;
    struct sockaddr_in addr;
    netq_marshal_t marshal;
    netq_demarshal_t demarshal;
    tcfree_fn frn;
    pthread_t thr;
    int security;
    char key[SHA_DIGEST_LENGTH];
} netq_t;

typedef struct {
    u_int32_t magic;
    u_int32_t op;
    char key[24];
    char data[65536];
} nq_packet;

#define NQ_MAGIC 0x4e455451

#define NQ_CONNECT       1
#define NQ_AUTH_REQUEST  2
#define NQ_AUTH_REPLY    3
#define NQ_AUTH_PASS     4
#define NQ_AUTH_FAIL     5
#define NQ_PACKET        6


static void *nq_recv(void *p);

static tclist_t *connections;

static void *
nq_send(void *p)
{
    netq_t *nq = p;
    nq_packet out;

    out.magic = htonl(NQ_MAGIC);
    out.op = htonl(NQ_PACKET);
    memcpy(out.key, nq->key, SHA_DIGEST_LENGTH);

    for(;;){
	int size;
	void *evt;

	evt = eventq_recv(nq->q);
	if(nq->marshal(evt, out.data, &size, &nq->addr)){
	    if(sendto(nq->socket, &out, size + offsetof(nq_packet, data),
		      0, (struct sockaddr *) &nq->addr, sizeof(nq->addr)) < 0){
		perror("sendto");
		break;
	    }
	}
	if(nq->frn)
	    nq->frn(evt);
    }

    return NULL;
}

extern int
nq_connect(eventq_t sq, netq_marshal_t marshal, tcfree_fn sfrn,
	   eventq_t rq, netq_demarshal_t demarshal, tcfree_fn rfrn,
	   struct sockaddr_in *addr, char *host, u_int16_t port,
	   int sec, char *key, int klen)
{
    netq_t *nq;
    nq_packet out;
    fd_set fds;
    struct timeval tmo = { .tv_sec = 1, .tv_usec = 0 };

    nq = calloc(1, sizeof(netq_t));
    nq->q = sq;
    nq->security = sec;
    nq->marshal = marshal;
    nq->frn = sfrn;

    nq->addr.sin_family = AF_INET;
    if(host){
	struct hostent *hostinfo;
	hostinfo = gethostbyname(host);
	if(hostinfo == NULL){
	    free(nq);
	    return NETQ_ENOHOST;
	}
	nq->addr.sin_addr = *(struct in_addr *) hostinfo->h_addr;
	nq->addr.sin_port = htons(port);
    } else {
	nq->addr = *addr;
    }

    if((nq->socket = socket(PF_INET, SOCK_DGRAM, 0)) < 0){
	free(nq);
	return NETQ_ESOCKET;
    }

    if(sec){
	SHA1(key, klen, nq->key);
    }

    out.magic = htonl(NQ_MAGIC);
    out.op = htonl(NQ_CONNECT);
    sendto(nq->socket, &out, offsetof(nq_packet, key),
	   0, (struct sockaddr *) &nq->addr, sizeof(nq->addr));

    FD_ZERO(&fds);
    FD_SET(nq->socket, &fds);
    if(select(FD_SETSIZE, &fds, NULL, NULL, &tmo) > 0){
	int s;
	nq_packet in;
	s = read(nq->socket, &in, sizeof(in));
	if(s < offsetof(nq_packet, key) || ntohl(in.magic) != NQ_MAGIC)
	    return NETQ_ESOCKET;
	in.op = ntohl(in.op);
	if(in.op == NQ_AUTH_REQUEST){
	    if(!key)
		return NETQ_EAUTHREQD;

	    out.op = htonl(NQ_AUTH_REPLY);
	    memcpy(out.key, nq->key, SHA_DIGEST_LENGTH);
	    sendto(nq->socket, &out, offsetof(nq_packet, data),
		   0, (struct sockaddr *) &nq->addr, sizeof(nq->addr));

	    FD_SET(nq->socket, &fds);
	    if(select(FD_SETSIZE, &fds, NULL, NULL, &tmo) <= 0)
		return NETQ_ESOCKET;

	    s = read(nq->socket, &in, sizeof(in));
	    if(s < offsetof(nq_packet, key) || ntohl(in.magic) != NQ_MAGIC)
		return NETQ_ESOCKET;

	    if(ntohl(in.op) != NQ_AUTH_PASS)
		return NETQ_EAUTHFAIL;
	} else if(in.op != NQ_AUTH_PASS){
	    return NETQ_EAUTHFAIL;
	}
    } else {
	return NETQ_ESOCKET;
    }

    nq->scntp = malloc(sizeof(int));
    *nq->scntp = 1;

    if(addr){
	int l = sizeof(*addr);
	getsockname(nq->socket, (struct sockaddr *) addr, &l);
    }

    tclist_push(connections, nq);
    pthread_create(&nq->thr, NULL, nq_send, nq);

    if(rq){
	netq_t *rnq = calloc(1, sizeof(netq_t));
	rnq->q = rq;
	rnq->security = sec;
	rnq->demarshal = demarshal;
	rnq->frn = rfrn;
	rnq->socket = nq->socket;
	rnq->scntp = nq->scntp;
	(*rnq->scntp)++;
	memcpy(rnq->key, nq->key, SHA_DIGEST_LENGTH);

	tclist_push(connections, rnq);
	pthread_create(&rnq->thr, NULL, nq_recv, rnq);
    }

    return 0;
}


static void *
nq_recv(void *p)
{
    netq_t *nq = p;
    nq_packet in, out;
    int size;
    struct sockaddr_in sa;
    socklen_t ads = sizeof(sa);

    out.magic = htonl(NQ_MAGIC);
    
    while((size = recvfrom(nq->socket, &in, sizeof(in), 0,
			   (struct sockaddr *) &sa, &ads)) > 0){
	void *evt;
	if(size < 8 || ntohl(in.magic) != NQ_MAGIC)
	    continue;

	in.op = ntohl(in.op);
	switch(in.op){
	case NQ_CONNECT:
/* 	    fprintf(stderr, "connect %s %i %i\n", inet_ntoa(sa.sin_addr), */
/* 		    ntohs(sa.sin_port), size); */
	    if(nq->security){
		out.op = htonl(NQ_AUTH_REQUEST);
	    } else {
		out.op = htonl(NQ_AUTH_PASS);
	    }
	    sendto(nq->socket, &out, offsetof(nq_packet, key), 0,
		   (struct sockaddr *) &sa, sizeof(sa));
	    break;

	case NQ_AUTH_REPLY:
	    if(nq->security){
		if(size < offsetof(nq_packet, key) + SHA_DIGEST_LENGTH){
		    fprintf(stderr, "Netq: Bad auth reply.\n");
		    break;
		}
		if(memcmp(in.key, nq->key, SHA_DIGEST_LENGTH) == 0){
		    out.op = htonl(NQ_AUTH_PASS);
		} else {
		    out.op = htonl(NQ_AUTH_FAIL);
		    fprintf(stderr, "Netq: Authorization failed from %s.\n",
			    inet_ntoa(sa.sin_addr));
		}
		sendto(nq->socket, &out, offsetof(nq_packet, key), 0,
		       (struct sockaddr *) &sa, sizeof(sa));
	    } else {
		fprintf(stderr,
			"%s: Auth reply received on non-authorized socket.\n",
			__FUNCTION__);
	    }
	    break;

	case NQ_PACKET:
	    if(nq->security && memcmp(in.key, nq->key, SHA_DIGEST_LENGTH))
		break;
	    if(nq->demarshal(in.data, &evt, size - offsetof(nq_packet, data),
			     &sa)){
		eventq_send(nq->q, evt);
		if(nq->frn)
		    nq->frn(evt);
	    }
	    break;
	}
    }
    perror("recvfrom");

    return NULL;
}

extern int
nq_listen(eventq_t q, struct sockaddr_in *addr, int sec, char *key, int klen,
	  netq_demarshal_t demarshal, tcfree_fn frn)
{
    netq_t *nq;
    int l;

    nq = malloc(sizeof(netq_t));
    nq->q = q;
    nq->demarshal = demarshal;
    nq->frn = frn;

    nq->security = sec;
    if(sec)
	SHA1(key, klen, nq->key);

    nq->addr.sin_family = AF_INET;
    nq->addr.sin_port = addr->sin_port;
    nq->addr.sin_addr = addr->sin_addr;

    if((nq->socket = socket(PF_INET, SOCK_DGRAM, 0)) < 0){
	free(nq);
	return -1;
    }

    if(bind(nq->socket, (struct sockaddr *) &nq->addr, sizeof(nq->addr)) < 0){
	close(nq->socket);
	free(nq);
	return -1;
    }

    nq->scntp = malloc(sizeof(int));
    *nq->scntp = 1;

    tclist_push(connections, nq);
    pthread_create(&nq->thr, NULL, nq_recv, nq);

    l = sizeof(*addr);
    getsockname(nq->socket, (struct sockaddr *) addr, &l);
    return 0;
}


static void
nq_stop(netq_t *nq)
{
    pthread_cancel(nq->thr);
    pthread_join(nq->thr, NULL);
    if(--*nq->scntp == 0){
	close(nq->socket);
	free(nq->scntp);
    }
    free(nq);
}

extern int
nq_close(eventq_t q)
{
    /* The first member of netq_t is the eventq_t. */
    int cmp(const void *p1, const void *p2){
	return *(void **)p1 - *(void **)p2;
    }

    tclist_delete(connections, &q, cmp, (tcfree_fn) nq_stop);
    return 0;
}


extern int
nq_init(char *opt)
{
    connections = tclist_new(TC_LOCK_SLOPPY);
    return 0;
}

extern int
nq_shutdown(void)
{
    tclist_destroy(connections, (tcfree_fn) nq_stop);
    return 0;
}
